

bits 16
org	0x9000

IDT_SEG		equ	0x9b00	; base address of temporary IDT
GDT_SEG		equ	0x9c00	; base address of temporary GDT
PAGE_TABLE1	equ	0x9d00	; page table mapped 0~4M
PAGE_TABLE2	equ	0x9e00	; page table mapped 4M~8M
PAGE_DIR	equ	0x9f00	; page DIR of kernel

KERNEL_SEG	equ	0x100	; segment address of KERNEL image 0x1000

	; make data segment equal to code segment
	mov	ax, cs
	mov	ds, ax


	; enable a20...
	call	enable_a20
	
	; enable protected mode...
	mov	ax, GDT_SEG
	mov	es, ax

	; source : ds:si = 0x9000:gdt
	; destination : es:di = 0x9c000:0x0
	; size = 0x18 * 2(word) = 0x30(48 bytes)
	xor	di, di
	mov	si, gdt
	xor	ecx, ecx
	mov	cx, 0x18
	rep	movsw
	


	; mask interrupts
	cli

	; load IDTR, GDTR
	lidt	[idtr]
	lgdt	[gdtr]
	
	; set system protected mode
	mov	eax, cr0
	or	al, 1			; set bit 0
	mov	cr0, eax
	
	; set segment selector index 3
	; 0x10 = 0001 0000 => index 00010(0x2), TI=0, RPL=00(0x0)
	mov	ax, 0x10
	mov	ds, ax
	mov	es, ax
	mov	fs, ax
	mov	gs, ax
	mov	ss, ax
	
	; start protected mode
	jmp	dword 0x8:in_protected

; enable a20...
enable_a20:
_enable_a20_1:
	in	al, 0x64
	test	al, 2
	jnz	_enable_a20_1
	mov	al, 0xd1
	out	0x64, al
	
_enable_a20_2:
	in	al, 0x64
	test	al, 2
	jnz	_enable_a20_2
	mov	al, 0xdf
	out	0x60, al
	ret


;----start 32bit protected mode----
bits 32

in_protected:


	; get total physical memory
	xor	edi, edi
	mov	al, 'Q'

	; write one byte data and read it,
	; if it read same value, there is a real memory.
	; but if read no value, it is end of memory.
	; not bad~~ ;-)
_1:	mov	[edi], al
	cmp	byte[edi], 'Q'
	jnz	_2
	add	edi, 0x100000		; 0x100000 = 2^20 = 1M
	jmp	_1
_2:	mov	edx, edi
	shr	edx, 20			; count * 2^20 = memory size in MB


	xor	ecx, ecx

	; setup page table... => directory
	; store value in ax at es:di(=0x10:0x9d000)

	; 0x9d000 => first page table (0~4M)
	; 0x9e000 => second page table (4M~8M)

	; clear memory from 0x9d000 ~ 0x9ffff (12K)
	; clear entries of page tables and page DIR
	mov	edi, 0x9d000
	xor	eax, eax		; store zeroes
	mov	cx, 3072		; size = 12K/4 = 3K
	rep	stosd
	
	;-----------------------
	; Build PAGE DIRECTORY
	;-----------------------
	; 0x93003 => base 0x93000, present, read&write
	; entry 0 and entry 0xc00 is mapping to physical memory area 0~4M
	; endty 1 and entry 0xc01 is mapping to physical memory area 4M~8M
	; This mapping makes KERNEL accessed by segment address after 3G
	mov	edi, 0x9f000
	
	; entry 0 = 0~4M table address is 0x9d000
	mov	eax, 0x9d003	
	stosd
	; entry 1 = 4M~8M, table address is 0x9e000
	mov	eax, 0x9e003
	stosd

	mov	edi, 0x9fc00
	
	; entry 0 = 0~4M table address is 0x9d000
	mov	eax, 0x9d003
	stosd
	; entry 1 = 4M~8M, table address is 0x9e000
	mov	eax, 0x9e003	
	stosd

	;-------------------------
	; Build PAGE TABLE
	;------------------------
	mov	edi, 0x9d000	; base address of table
	xor	eax, eax
	or	eax, 0x3		; set attributes, present, r&w
	
	; write table entry
	; map 0~4M-1
	mov	cx, 1024		; count of table entry is 1024
_3:	stosd				; store base address of page frame
	add	eax, 0x1000		; next page frame (0x1000 = 4K = 1 page)
	dec	cx
	jnz	_3

	; clear attributes, set new attributes, present, read only
	and	eax, 0xfffff000
	or	eax, 0x3		; present, read only
	
	
	; write table entry
	; 4M~8M-1
	mov	cx, 1024
_4:	stosd
	add	eax, 0x1000
	dec	cx
	jnz	_4

	; store base address of page directory
	mov	eax, 0x9f000
	mov	cr3, eax

	; set paging bit
	mov	eax, cr0
	or	eax, 0x80000000
	mov	cr0, eax


	; jump to kernel
	; 0xc0001000 => 0x001000(4K) (way of paging)
	jmp 0xc0001000


;--------------------------
; GLOBAL DESCRIPTOR TABLE
;-------------------------
gdt:	dw 0, 0, 0, 0

		dw 0xffff, 0, 0x9a00, 0xcf
		dw 0xffff, 0, 0x9200, 0xcf

		dw 0, 0, 0, 0	; dummy
		
		dw 0xffff, 0, 0xfa00, 0xcf	
		dw 0xffff, 0, 0xf200, 0xcf

; GDTR value
gdtr:	dw 0x1000
		dd 0x9c000

; IDTR value
idtr:	dw 0x1000
		dd 0x9b000

	; make Setup.asm to 512 byte size
	times 512-($-$$) db 0
